#ifndef streamer_h__
#define streamer_h__

/*
                Streamer
                ========

    Copyright (c) 20010-2011 Khaled Alshaya

    Distributed under the Boost Software License, version 1.0
    (See the license at: http://www.boost.org/license_1_0.txt).
*/

#include <cstddef>
#include <type_traits>
#include <ostream>
#include <utility>
#include <tuple>

namespace streamer{
    namespace detail{
        template <class T> struct is_char{ static const bool value = false; };
        template <> struct is_char<char>{ static const bool value = true; };
        template <> struct is_char<wchar_t>{ static const bool value = true; };
        template <> struct is_char<const char>{ static const bool value = true; };
        template <> struct is_char<const wchar_t>{ static const bool value = true; };
        template <class T>
        struct is_sequence{
            typedef char yes[1];
            typedef char no[2];
            template <class C>
            static yes& test(typename C::allocator_type*);
            template <class>
            static no& test(...);
            static constexpr bool value = sizeof(test<T>(nullptr)) == sizeof(yes);
        };
        template <class T>
        struct is_associative{
            typedef char yes[1];
            typedef char no[2];
            template <class C>
            static yes& test(typename C::key_type*);
            template <class>
            static no& test(...);
            static constexpr bool value = sizeof(test<T>(nullptr)) == sizeof(yes);
        };
        template <class T>
        struct is_adaptor{
            typedef char yes[1];
            typedef char no[2];
            template <class C>
            static yes& test(typename C::container_type*);
            template <class>
            static no& test(...);
            static constexpr bool value = sizeof(test<T>(nullptr)) == sizeof(yes);
        };
        template <class charT, class traits, class Tuple, std::size_t index>
        struct tuple_printer{
            static void print(std::basic_ostream<charT, traits>& out, const Tuple& tuple){
                tuple_printer<charT, traits, Tuple, index-1>::print(out, tuple);
                out << ", " << std::get<index-1>(tuple);
            }
        };
        template <class charT, class traits, class Tuple>
        struct tuple_printer<charT, traits, Tuple, 1>{
            static void print(std::basic_ostream<charT, traits>& out, const Tuple& tuple){
                out << std::get<0>(tuple);
            }
        };

        template <class Adaptor>
        const typename Adaptor::container_type& container(const Adaptor& adaptor){
            struct HackedAdaptor : private Adaptor{
                static const typename Adaptor::container_type& container(const Adaptor& adaptor){
                    return adaptor.*&HackedAdaptor::c;
                }
            };
            return HackedAdaptor::container(adaptor);
        }
        template <class charT, class traits, class Sequence>
        std::basic_ostream<charT, traits>&
        stream_sequence(std::basic_ostream<charT, traits>& out, const Sequence& sequence){
            out << "[ ";
            for(const auto& element : sequence) out << element << ' ';
            return out << ']';
        }
        template <class charT, class traits, class Associative>
        std::basic_ostream<charT, traits>&
        stream_associative(std::basic_ostream<charT, traits>& out, const Associative& associative){
            out << "{ ";
            for(const auto& element : associative) out << element << ' ';
            return out << '}';
        }
    }

    template <class charT, class traits, class T, std::size_t N>
    typename std::enable_if<!detail::is_char<T>::value, std::basic_ostream<charT, traits>&>::type
    operator<<(std::basic_ostream<charT, traits>& out, T (&array)[N]){
        return detail::stream_sequence(out, array);
    }
    template <class charT, class traits, class T1, class T2>
    std::basic_ostream<charT, traits>&
    operator<<(std::basic_ostream<charT, traits>& out, const std::pair<T1, T2>& pair){
        return out << '(' << pair.first << ',' << pair.second << ')';
    }
    template <class charT, class traits, class... Types>
    std::basic_ostream<charT, traits>&
    operator<<(std::basic_ostream<charT, traits>& out, const std::tuple<Types...>& tuple){
        constexpr std::size_t tuple_size = sizeof...(Types);
        out << "(";
        detail::tuple_printer<charT, traits, std::tuple<Types...>, tuple_size>::print(out, tuple);
        return out << ")";
    }
    template <class charT, class traits, class T, template <class> class InitializerList>
    std::basic_ostream<charT, traits>&
    operator<<(std::basic_ostream<charT, traits>& out, const InitializerList<T>& initializer_list){
        return detail::stream_sequence(out, initializer_list);
    }
    template <class charT, class traits, class T, std::size_t N, template <class, std::size_t> class Array>
    std::basic_ostream<charT, traits>&
    operator<<(std::basic_ostream<charT, traits>& out, const Array<T, N>& array){
        return detail::stream_sequence(out, array);
    }
    template <class charT, class traits, class Sequence>
    typename std::enable_if<detail::is_sequence<Sequence>::value && !detail::is_associative<Sequence>::value, std::basic_ostream<charT, traits>&>::type
    operator<<(std::basic_ostream<charT, traits>& out, const Sequence& sequence){
        return detail::stream_sequence(out, sequence);
    }
    template <class charT, class traits, class Associative>
    typename std::enable_if<detail::is_associative<Associative>::value, std::basic_ostream<charT, traits>&>::type
    operator<<(std::basic_ostream<charT, traits>& out, const Associative& associative){
        return detail::stream_associative(out, associative);
    }
    template <class charT, class traits, class Adaptor>
    typename std::enable_if<detail::is_adaptor<Adaptor>::value, std::basic_ostream<charT, traits>&>::type
    operator<<(std::basic_ostream<charT, traits>& out, const Adaptor& adaptor){
        return detail::stream_sequence(out, detail::container(adaptor));
    }
}

#endif